; crc32.asm -- compute the CRC-32 of a data stream
; Copyright (C) 1995-2006, 2010, 2011, 2012 Mark Adler
; For conditions of distribution and use, see copyright notice in zlib.inc

; Thanks to Rodney Brown <rbrown64@csc.com.au> for his contribution of faster
; CRC methods: exclusive-oring 32 bits of data at a time, and pre-computing
; tables for updating the shift register in one step with three exclusive-ors
; instead of four steps with four exclusive-ors.  This results in about a
; factor of two increase in speed on a Power PC G4 (PPC7455) using gcc -O3.


;  Note on the use of DYNAMIC_CRC_TABLE: there is no mutex or semaphore
;  protection on the static variables used to control the first-use generation
;  of the crc tables.  Therefore, if you #define DYNAMIC_CRC_TABLE, you should
;  first call get_crc_table() to initialize the tables before allowing more than
;  one thread to use crc32().

;  DYNAMIC_CRC_TABLE and MAKECRCH can be #defined to write out crc32.inc.


; Definitions for doing the crc four data bytes at a time.

BYFOUR equ 0
MAKECRCH equ 0
DYNAMIC_CRC_TABLE equ 1

if BYFOUR eq 1
	TBLS equ 8
else
	TBLS equ 1
end if ;BYFOUR

if DYNAMIC_CRC_TABLE eq 1

align 4
crc_table_empty dd 1
align 4
crc_table rd TBLS*256
;local void make_crc_table OF((void));

;  Generate tables for a byte-wise 32-bit CRC calculation on the polynomial:
;  x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x+1.

;  Polynomials over GF(2) are represented in binary, one bit per coefficient,
;  with the lowest powers in the most significant bit.  Then adding polynomials
;  is just exclusive-or, and multiplying a polynomial by x is a right shift by
;  one.  If we call the above polynomial p, and represent a byte as the
;  polynomial q, also with the lowest power in the most significant bit (so the
;  byte 0xb1 is the polynomial x^7+x^3+x+1), then the CRC is (q*x^32) mod p,
;  where a mod b means the remainder after dividing a by b.

;  This calculation is done using the shift-register method of multiplying and
;  taking the remainder.  The register is initialized to zero, and for each
;  incoming bit, x^32 is added mod p to the register if the bit is a one (where
;  x^32 mod p is p+x^32 = x^26+...+1), and the register is multiplied mod p by
;  x (which is shifting right by one and adding x^32 mod p if the bit shifted
;  out is a one).  We start with the highest power (least significant bit) of
;  q and repeat for all eight bits of q.

;  The first table is simply the CRC of all possible eight bit values.  This is
;  all the information needed to generate CRCs on data a byte at a time for all
;  combinations of CRC register values and incoming bytes.  The remaining tables
;  allow for word-at-a-time CRC calculation for both big-endian and little-
;  endian machines, where a word is four bytes.

;void ()
align 4
proc make_crc_table uses ecx edx edi
;    z_crc_t c;
;    int n, k;
;    z_crc_t poly ; polynomial exclusive-or pattern
	; terms of polynomial defining this crc (except x^32):
;    int first = 1;      /* flag to limit concurrent making */
;    const unsigned char p[] = {0,1,2,4,5,7,8,10,11,12,16,22,23,26};

	; See if another task is already doing this (not thread-safe, but better
	; than nothing -- significantly reduces duration of vulnerability in
	; case the advice about DYNAMIC_CRC_TABLE is ignored)
;    if (first) {
;        first = 0;

	; make exclusive-or pattern from polynomial (0xedb88320)
;        poly = 0;
;        for (n = 0; n < (int)(sizeof(p)/sizeof(unsigned char)); n++)
;            poly |= (z_crc_t)1 << (31 - p[n]);

	; generate a crc for every 8-bit value
	xor edx, edx
	mov edi, crc_table
.1:
	mov ecx, 8
	mov eax, edx
.2:
	shr eax, 1
	jnc @f
	xor eax, 0xEDB88320
@@:
	loop .2
	stosd
	inc dl
	jnz .1
;        for (n = 0; n < 256; n++) {
;            c = (z_crc_t)n;
;            for (k = 0; k < 8; k++)
;                c = c & 1 ? poly ^ (c >> 1) : c >> 1;
;            crc_table[0][n] = c;
;        }

if BYFOUR eq 1
	; generate crc for each value followed by one, two, and three zeros,
	; and then the byte reversal of those as well as the first table
;        for (n = 0; n < 256; n++) {
;            c = crc_table[0][n];
;            crc_table[4][n] = ZSWAP32(c);
;            for (k = 1; k < 4; k++) {
;                c = crc_table[0][c & 0xff] ^ (c >> 8);
;                crc_table[k][n] = c;
;                crc_table[k + 4][n] = ZSWAP32(c);
;            }
;        }
end if ;BYFOUR

		mov dword[crc_table_empty],0
;    }
;    else {      /* not first */
	; wait for the other guy to finish (not efficient, but rare)
;        while (crc_table_empty){};
;    }

if MAKECRCH eq 1
	; write out CRC tables to crc32.inc
;        FILE *out;

;        out = fopen("crc32.inc", "w");
;        if (out == NULL) return;
;        fprintf(out, "/* crc32.inc -- tables for rapid CRC calculation\n");
;        fprintf(out, " * Generated automatically by crc32.asm\n */\n\n");
;        fprintf(out, "local const z_crc_t FAR ");
;        fprintf(out, "crc_table[TBLS][256] =\n{\n  {\n");
;        write_table(out, crc_table[0]);
if BYFOUR eq 1
;        fprintf(out, "#ifdef BYFOUR\n");
;        for (k = 1; k < 8; k++) {
;            fprintf(out, "  },\n  {\n");
;            write_table(out, crc_table[k]);
;        }
;        fprintf(out, "#endif\n");
end if ;BYFOUR
;        fprintf(out, "  }\n};\n");
;        fclose(out);
end if ;MAKECRCH
	ret
endp

;void (out, table)
;    FILE *out;
;    const z_crc_t FAR *table;
align 4
proc write_table, p1out:dword, table:dword
;    int n;

;    for (n = 0; n < 256; n++)
;        fprintf(p1out, "%s0x%08lxUL%s", n % 5 ? "" : "    ",
;                (unsigned long)(table[n]),
;                n == 255 ? "\n" : (n % 5 == 4 ? ",\n" : ", "));
	ret
endp

else ;!DYNAMIC_CRC_TABLE
; ========================================================================
; Tables of CRC-32s of all single-byte values, made by make_crc_table().

;include 'crc32.inc'
end if ;DYNAMIC_CRC_TABLE

; =========================================================================
; This function can be used by asm versions of crc32()

;const z_crc_t FAR * ()
align 4
proc get_crc_table
if DYNAMIC_CRC_TABLE eq 1
	cmp dword[crc_table_empty],0
	je @f ;if (..)
		call make_crc_table
	@@:
end if ;DYNAMIC_CRC_TABLE
	mov eax,crc_table
	ret
endp

; =========================================================================
macro DO1
{
	xor al,byte[esi]
	xor al,ah
	mov eax,[crc_table+eax*4]
	inc esi
}
macro DO8
{
	DO1
	DO1
	DO1
	DO1
	DO1
	DO1
	DO1
	DO1
}

; =========================================================================
;unsigned long (crc, buf, len)
;    unsigned long crc;
;    const unsigned char FAR *buf;
;    uInt len;
align 4
proc calc_crc32 uses ecx, p1crc:dword, buf:dword, len:dword
	xor eax,eax
	cmp dword[buf],Z_NULL
	je .end_f ;if (..==0) return 0

if DYNAMIC_CRC_TABLE eq 1
	cmp dword[crc_table_empty],0
	je @f ;if (..)
		call make_crc_table
	@@:
end if ;DYNAMIC_CRC_TABLE

if BYFOUR eq 1
;    if (sizeof(void *) == sizeof(ptrdiff_t)) {
;        z_crc_t endian;

;        endian = 1;
;        if (*((unsigned char *)(&endian)))
;            return crc32_little(p1crc, buf, len);
;        else
;            return crc32_big(p1crc, buf, len);
;    }
end if ;BYFOUR
	mov eax,[p1crc]
	xor eax,0xffffffff
	mov [p1crc],eax
	mov ecx,[len]
align 4
	.cycle0:
	cmp ecx,8
	jl @f
		DO8
		sub ecx,8
		jmp .cycle0
align 4
	@@:
	cmp ecx,1
	jl @f
		DO1
		dec ecx
		jmp @b
	@@:
	mov eax,[p1crc]
	xor eax,0xffffffff
.end_f:
	ret
endp

if BYFOUR eq 1

; =========================================================================
;#define DOLIT4 c ^= *buf4++; \
;        c = crc_table[3][c & 0xff] ^ crc_table[2][(c >> 8) & 0xff] ^ \
;            crc_table[1][(c >> 16) & 0xff] ^ crc_table[0][c >> 24]
macro DOLIT32
{
	DOLIT4
	DOLIT4
	DOLIT4
	DOLIT4
	DOLIT4
	DOLIT4
	DOLIT4
	DOLIT4
}

; =========================================================================
;unsigned long (crc, buf, len)
;    unsigned long crc;
;    const unsigned char FAR *buf;
;    unsigned len;
align 4
proc crc32_little, crc:dword, buf:dword, len:dword
;    register z_crc_t c;
;    register const z_crc_t FAR *buf4;

;    c = (z_crc_t)crc;
;    c = ~c;
;    while (len && ((ptrdiff_t)buf & 3)) {
;        c = crc_table[0][(c ^ *buf++) & 0xff] ^ (c >> 8);
;        len--;
;    }

;    buf4 = (const z_crc_t FAR *)(const void FAR *)buf;
;    while (len >= 32) {
;        DOLIT32;
;        len -= 32;
;    }
;    while (len >= 4) {
;        DOLIT4;
;        len -= 4;
;    }
;    buf = (const unsigned char FAR *)buf4;

;    if (len) do {
;        c = crc_table[0][(c ^ *buf++) & 0xff] ^ (c >> 8);
;    } while (--len);
;    c = ~c;
;    return (unsigned long)c;
	ret
endp

; =========================================================================
;#define DOBIG4 c ^= *++buf4; \
;        c = crc_table[4][c & 0xff] ^ crc_table[5][(c >> 8) & 0xff] ^ \
;            crc_table[6][(c >> 16) & 0xff] ^ crc_table[7][c >> 24]
;#define DOBIG32 DOBIG4; DOBIG4; DOBIG4; DOBIG4; DOBIG4; DOBIG4; DOBIG4; DOBIG4

; =========================================================================
;unsigned long (crc, buf, len)
;    unsigned long crc;
;    const unsigned char FAR *buf;
;    unsigned len;
align 4
proc crc32_big, crc:dword, buf:dword, len:dword
;    register z_crc_t c;
;    register const z_crc_t FAR *buf4;

;    c = ZSWAP32((z_crc_t)crc);
;    c = ~c;
;    while (len && ((ptrdiff_t)buf & 3)) {
;        c = crc_table[4][(c >> 24) ^ *buf++] ^ (c << 8);
;        len--;
;    }

;    buf4 = (const z_crc_t FAR *)(const void FAR *)buf;
;    buf4--;
;    while (len >= 32) {
;        DOBIG32;
;        len -= 32;
;    }
;    while (len >= 4) {
;        DOBIG4;
;        len -= 4;
;    }
;    buf4++;
;    buf = (const unsigned char FAR *)buf4;

;    if (len) do {
;        c = crc_table[4][(c >> 24) ^ *buf++] ^ (c << 8);
;    } while (--len);
;    c = ~c;
;    return (unsigned long)(ZSWAP32(c));
	ret
endp

end if ;BYFOUR

;#define GF2_DIM 32      /* dimension of GF(2) vectors (length of CRC) */

; =========================================================================
;unsigned long (mat, vec)
;    unsigned long *mat;
;    unsigned long vec;
align 4
proc gf2_matrix_times, mat:dword, vec:dword
;    unsigned long sum;

;    sum = 0;
;    while (vec) {
;        if (vec & 1)
;            sum ^= *mat;
;        vec >>= 1;
;        mat++;
;    }
;    return sum;
	ret
endp

; =========================================================================
;local void (square, mat)
;    unsigned long *square;
;    unsigned long *mat;
align 4
proc gf2_matrix_square, square:dword, mat:dword
;    int n;

;    for (n = 0; n < GF2_DIM; n++)
;        square[n] = gf2_matrix_times(mat, mat[n]);
	ret
endp

; =========================================================================
;uLong (crc1, crc2, len2)
;    uLong crc1;
;    uLong crc2;
;    z_off64_t len2;
align 4
proc crc32_combine_, crc1:dword, crc2:dword, len2:dword
;    int n;
;    unsigned long row;
;    unsigned long even[GF2_DIM];    /* even-power-of-two zeros operator */
;    unsigned long odd[GF2_DIM];     /* odd-power-of-two zeros operator */

;    /* degenerate case (also disallow negative lengths) */
;    if (len2 <= 0)
;        return crc1;

;    /* put operator for one zero bit in odd */
;    odd[0] = 0xedb88320UL;          /* CRC-32 polynomial */
;    row = 1;
;    for (n = 1; n < GF2_DIM; n++) {
;        odd[n] = row;
;        row <<= 1;
;    }

;    /* put operator for two zero bits in even */
;    gf2_matrix_square(even, odd);

;    /* put operator for four zero bits in odd */
;    gf2_matrix_square(odd, even);

;    /* apply len2 zeros to crc1 (first square will put the operator for one
;       zero byte, eight zero bits, in even) */
;    do {
;        /* apply zeros operator for this bit of len2 */
;        gf2_matrix_square(even, odd);
;        if (len2 & 1)
;            crc1 = gf2_matrix_times(even, crc1);
;        len2 >>= 1;

;        /* if no more bits set, then done */
;        if (len2 == 0)
;            break;

;        /* another iteration of the loop with odd and even swapped */
;        gf2_matrix_square(odd, even);
;        if (len2 & 1)
;            crc1 = gf2_matrix_times(odd, crc1);
;        len2 >>= 1;

	; if no more bits set, then done
;    } while (len2 != 0);

	; return combined crc
;    crc1 ^= crc2;
;    return crc1;
	ret
endp

; =========================================================================
;uLong (crc1, crc2, len2)
;    uLong crc1;
;    uLong crc2;
;    z_off_t len2;
align 4
proc crc32_combine, crc1:dword, crc2:dword, len2:dword
	stdcall crc32_combine_, [crc1], [crc2], [len2]
	ret
endp

;uLong (crc1, crc2, len2)
;    uLong crc1;
;    uLong crc2;
;    z_off64_t len2;
align 4
proc crc32_combine64, crc1:dword, crc2:dword, len2:dword
	stdcall crc32_combine_, [crc1], [crc2], [len2]
	ret
endp
